// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:copy_with_extension/copy_with_extension.dart';
import 'package:flutter/material.dart';
import 'package:wyatt_ui_components/src/core/utils/component_registry.dart';
import 'package:wyatt_ui_components/src/domain/entities/entities.dart';
part 'component_theme_data.g.dart';

/// {@template component_theme_data}
/// A class that holds all the components that are used in the app.
/// {@endtemplate}
@CopyWith()
class ComponentThemeData {
  /// {@macro component_theme_data}
  factory ComponentThemeData({
    ComponentRegistry<TopAppBarComponent>? topAppBars,
    ComponentRegistry<TopNavigationBarComponent>? topNavigationBars,
    ComponentRegistry<BottomNavigationBarComponent>? bottomNavigationBars,
    ComponentRegistry<ErrorComponent>? errors,
    ComponentRegistry<LoaderComponent>? loaders,
    ComponentRegistry<RichTextBuilderComponent>? richTextBuilders,
    ComponentRegistry<TextInputComponent>? textInputs,
    ComponentRegistry<FileSelectionButtonComponent>? fileSelectionButtons,
    ComponentRegistry<FlatButtonComponent>? flatButtons,
    ComponentRegistry<SimpleIconButtonComponent>? simpleIconButtons,
    ComponentRegistry<SymbolButtonComponent>? symbolButtons,
    ComponentRegistry<InformationCardComponent>? informationCards,
    ComponentRegistry<PortfolioCardComponent>? portfolioCards,
    ComponentRegistry<QuoteCardComponent>? quoteCards,
    ComponentRegistry<SkillCardComponent>? skillCards,
    ComponentRegistry<PricingCardComponent>? pricingCards,
    ComponentRegistry<FloatingActionButtonComponent>? floatingActionButtons,
  }) =>
      ComponentThemeData.raw(
        topAppBars: topAppBars,
        topNavigationBars: topNavigationBars,
        bottomNavigationBars: bottomNavigationBars,
        errors: errors,
        loaders: loaders,
        richTextBuilders: richTextBuilders,
        textInputs: textInputs,
        fileSelectionButtons: fileSelectionButtons,
        flatButtons: flatButtons,
        simpleIconButtons: simpleIconButtons,
        symbolButtons: symbolButtons,
        informationCards: informationCards,
        portfolioCards: portfolioCards,
        quoteCards: quoteCards,
        skillCards: skillCards,
        pricingCards: pricingCards,
        floatingActionButtons: floatingActionButtons,
      );

  // /// {@macro component_theme_data}
  factory ComponentThemeData.fromOther(ComponentThemeData other) =>
      ComponentThemeData(
        topAppBars: other._topAppBars,
        topNavigationBars: other._topNavigationBars,
        bottomNavigationBars: other._bottomNavigationBars,
        errors: other._errors,
        loaders: other._loaders,
        richTextBuilders: other._richTextBuilders,
        textInputs: other._textInputs,
        fileSelectionButtons: other._fileSelectionButtons,
        flatButtons: other._flatButtons,
        simpleIconButtons: other._simpleIconButtons,
        symbolButtons: other._symbolButtons,
        informationCards: other._informationCards,
        portfolioCards: other._portfolioCards,
        quoteCards: other._quoteCards,
        skillCards: other._skillCards,
        pricingCards: other._pricingCards,
        floatingActionButtons: other._floatingActionButtons,
      );

  /// Create a [ComponentThemeData] given a set of exact values. Most values
  /// must be specified.
  ///
  /// This will rarely be used directly. It is used by lerp to
  /// create intermediate themes based on two themes created with the
  /// [ComponentThemeData] constructor.
  const ComponentThemeData.raw({
    ComponentRegistry<TopAppBarComponent>? topAppBars,
    ComponentRegistry<TopNavigationBarComponent>? topNavigationBars,
    ComponentRegistry<BottomNavigationBarComponent>? bottomNavigationBars,
    ComponentRegistry<ErrorComponent>? errors,
    ComponentRegistry<LoaderComponent>? loaders,
    ComponentRegistry<RichTextBuilderComponent>? richTextBuilders,
    ComponentRegistry<TextInputComponent>? textInputs,
    ComponentRegistry<FileSelectionButtonComponent>? fileSelectionButtons,
    ComponentRegistry<FlatButtonComponent>? flatButtons,
    ComponentRegistry<SimpleIconButtonComponent>? simpleIconButtons,
    ComponentRegistry<SymbolButtonComponent>? symbolButtons,
    ComponentRegistry<InformationCardComponent>? informationCards,
    ComponentRegistry<PortfolioCardComponent>? portfolioCards,
    ComponentRegistry<QuoteCardComponent>? quoteCards,
    ComponentRegistry<SkillCardComponent>? skillCards,
    ComponentRegistry<PricingCardComponent>? pricingCards,
    ComponentRegistry<FloatingActionButtonComponent>? floatingActionButtons,
  })  : _floatingActionButtons = floatingActionButtons,
        _symbolButtons = symbolButtons,
        _simpleIconButtons = simpleIconButtons,
        _flatButtons = flatButtons,
        _fileSelectionButtons = fileSelectionButtons,
        _textInputs = textInputs,
        _richTextBuilders = richTextBuilders,
        _pricingCards = pricingCards,
        _skillCards = skillCards,
        _quoteCards = quoteCards,
        _portfolioCards = portfolioCards,
        _informationCards = informationCards,
        _loaders = loaders,
        _errors = errors,
        _bottomNavigationBars = bottomNavigationBars,
        _topNavigationBars = topNavigationBars,
        _topAppBars = topAppBars;

  R _get<T extends Component, R>(T? component, R? returned) {
    if (component == null) {
      throw FlutterError('No $T component provided.\n'
          'Please provide a $T component in the ComponentThemeData.');
    }

    if (returned == null) {
      throw FlutterError('$T does not have a CopyWith method.\n'
          'Please provide a $T component that implements CopyWith '
          'in the ComponentThemeData.');
    }

    return returned;
  }

  // Bars
  final ComponentRegistry<TopAppBarComponent>? _topAppBars;
  $TopAppBarComponentCWProxy topAppBarComponent([String? id]) =>
      _get(_topAppBars?.call(id), _topAppBars?.call(id)?.copyWith);

  final ComponentRegistry<TopNavigationBarComponent>? _topNavigationBars;
  $TopNavigationBarComponentCWProxy topNavigationBarComponent([String? id]) =>
      _get(
        _topNavigationBars?.call(id),
        _topNavigationBars?.call(id)?.copyWith,
      );

  final ComponentRegistry<BottomNavigationBarComponent>? _bottomNavigationBars;
  $BottomNavigationBarComponentCWProxy bottomNavigationBarComponent([
    String? id,
  ]) =>
      _get(
        _bottomNavigationBars?.call(id),
        _bottomNavigationBars?.call(id)?.copyWith,
      );

  // CRUD Widgets
  final ComponentRegistry<ErrorComponent>? _errors;
  $ErrorComponentCWProxy errorComponent([String? id]) =>
      _get(_errors?.call(id), _errors?.call(id)?.copyWith);

  final ComponentRegistry<LoaderComponent>? _loaders;
  $LoaderComponentCWProxy loaderComponent([String? id]) =>
      _get(_loaders?.call(id), _loaders?.call(id)?.copyWith);

  // Cards
  final ComponentRegistry<InformationCardComponent>? _informationCards;
  $InformationCardComponentCWProxy informationCardComponent([String? id]) =>
      _get(_informationCards?.call(id), _informationCards?.call(id)?.copyWith);

  final ComponentRegistry<PortfolioCardComponent>? _portfolioCards;
  $PortfolioCardComponentCWProxy portfolioCardComponent([String? id]) =>
      _get(_portfolioCards?.call(id), _portfolioCards?.call(id)?.copyWith);

  final ComponentRegistry<QuoteCardComponent>? _quoteCards;
  $QuoteCardComponentCWProxy quoteCardComponent([String? id]) =>
      _get(_quoteCards?.call(id), _quoteCards?.call(id)?.copyWith);

  final ComponentRegistry<SkillCardComponent>? _skillCards;
  $SkillCardComponentCWProxy skillCardComponent([String? id]) =>
      _get(_skillCards?.call(id), _skillCards?.call(id)?.copyWith);
  final ComponentRegistry<PricingCardComponent>? _pricingCards;
  $PricingCardComponentCWProxy pricingCardComponent([String? id]) =>
      _get(_pricingCards?.call(id), _pricingCards?.call(id)?.copyWith);

  // Rich Text
  final ComponentRegistry<RichTextBuilderComponent>? _richTextBuilders;
  $RichTextBuilderComponentCWProxy richTextBuilderComponent([String? id]) =>
      _get(_richTextBuilders?.call(id), _richTextBuilders?.call(id)?.copyWith);

  // Text Inputs
  final ComponentRegistry<TextInputComponent>? _textInputs;
  $TextInputComponentCWProxy textInputComponent([String? id]) =>
      _get(_textInputs?.call(id), _textInputs?.call(id)?.copyWith);

  // Buttons
  final ComponentRegistry<FileSelectionButtonComponent>? _fileSelectionButtons;
  $FileSelectionButtonComponentCWProxy fileSelectionButtonComponent([
    String? id,
  ]) =>
      _get(
        _fileSelectionButtons?.call(id),
        _fileSelectionButtons?.call(id)?.copyWith,
      );

  final ComponentRegistry<FlatButtonComponent>? _flatButtons;
  $FlatButtonComponentCWProxy flatButtonComponent([String? id]) =>
      _get(_flatButtons?.call(id), _flatButtons?.call(id)?.copyWith);

  final ComponentRegistry<SimpleIconButtonComponent>? _simpleIconButtons;
  $SimpleIconButtonComponentCWProxy simpleIconButtonComponent([String? id]) =>
      _get(
        _simpleIconButtons?.call(id),
        _simpleIconButtons?.call(id)?.copyWith,
      );

  final ComponentRegistry<SymbolButtonComponent>? _symbolButtons;
  $SymbolButtonComponentCWProxy symbolButtonComponent([String? id]) =>
      _get(_symbolButtons?.call(id), _symbolButtons?.call(id)?.copyWith);

  final ComponentRegistry<FloatingActionButtonComponent>?
      _floatingActionButtons;
  $FloatingActionButtonComponentCWProxy floatingActionButtonComponent([
    String? id,
  ]) =>
      _get(
        _floatingActionButtons?.call(id),
        _floatingActionButtons?.call(id)?.copyWith,
      );
}
